//
// Copyright (C) 2008-2009 Reto Gantenbein
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//

#include "MsgHandler.h"
#include "PacketProxy.h" 			//TODO: remove! only for debugging

Define_Module(MsgHandler);

MsgHandler::MsgHandler()
{
	host = NULL;
	for (int i = 0; i < MAX_NUM_VIF; i++)
		vif[i] = NULL;
	memset(&addrInfo, 0, sizeof(struct addrinfo));
	memset(&hostAddr, 0, sizeof(struct sockaddr_storage));
	numVif = 0;
}

MsgHandler::~MsgHandler() {}

void MsgHandler::initialize()
{
	host = check_and_cast<cModule *>(getParentModule());
	numVif = host->par("numVif");
	for (int i = 0; i < numVif; i++)
		vif[i] = check_and_cast<cModule *>(host->getSubmodule("ath", i));

	std::string hostIP = host->par("ipAddress");
	const char *ipp = hostIP.c_str();

	struct addrinfo request, *result ;
	memset(&request, 0, sizeof(struct addrinfo));
	request.ai_family = AF_UNSPEC;
	request.ai_socktype = SOCK_DGRAM;
	request.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
	request.ai_protocol = 0;
	request.ai_canonname = NULL;
	request.ai_addr = NULL;
	request.ai_next = NULL;
	int port = host->par("port");
    char cport[5];
    sprintf(cport, "%i", port);
    if (getaddrinfo(ipp, cport, &request, &result) < 0)
    	throw new cRuntimeError("MsgHandler: getaddrinfo");
    else {
    	memmove(&addrInfo, result, sizeof(struct addrinfo));
    	memmove(&hostAddr, result->ai_addr, sizeof(struct sockaddr_storage));
    	addrInfo.ai_addr = (struct sockaddr *) &hostAddr;
    }
}

void MsgHandler::handleMessage(cMessage *etherFrame)
{
	ev << "MsgHandler.handleMessage(): msg received!\n";   //TODO: for debug only!

	// we should only receive messages from lower layer
	RAWEtherFrame *frame = check_and_cast<RAWEtherFrame *>(etherFrame);

    int payloadSize = frame->getRawBytesArraySize();
    int messageSize = DATA_PREAMBLE_SIZE + payloadSize;

    // create wlan_msg
    dataMsg *msg = (dataMsg *) calloc(1, messageSize);
    ((msgHeader *) msg)->type = DATA_MSG;
    strcpy(((msgHeader *) msg)->id, "0\0");		//TODO: WlanModel id?
    char *payloadPointer = (char *) &(msg->wlanPkt);
    for (int i = 0; i < payloadSize; i++)
    {
        *payloadPointer = frame->getRawBytes(i);
        payloadPointer++;
    }
    msg->pktLength = payloadSize;

    // send back to host
    PacketProxy::printDataMessage((char *) msg);
    sendToHost((char *) msg, messageSize);

    // log packet processing time
	timeval curTime;
	gettimeofday(&curTime, NULL);
	timeval processingTimeval = timeval_substract(curTime, (frame->getTimestamp()).dbl());
	double processingTime = (processingTimeval.tv_sec + processingTimeval.tv_usec * 1e-6) * 1000;  // in ms
	ev << "PacketProxy.handleReply(): processingTime " << processingTime << "ms\n";
	//pktProcTime.record(processingTime);				//TODO: add logging

    delete frame;
    free(msg);
}

void MsgHandler::sendAck(unsigned int msgId)
{
	Enter_Method("sendAck()");
    plainMsg *ack;

    ack = (plainMsg *) calloc(1, ACK_MSG_SIZE);
    ack->msgId = msgId;
    ((msgHeader *) ack)->type = ACK_MSG;

//    debugMessage((char *) ack, ACK_MSG_SIZE);
    PacketProxy::printPlainMessage((char *) ack);			//TODO: only for debug!

    sendToHost((char *) ack, ACK_MSG_SIZE);

    free(ack);
}

void MsgHandler::sendToHost(char *msg, int length)
{
	cUDPSocketRTScheduler *rtScheduler =
		check_and_cast<cUDPSocketRTScheduler *>(simulation.getScheduler());
	rtScheduler->sendMessage(msg, length, (struct sockaddr *) &hostAddr,
							 addrInfo.ai_addrlen);
}

void MsgHandler::sendToModel(char *wlanPkt, int length)
{
	Enter_Method("sendToModel()");
    RAWEtherFrame *etherFrame = new RAWEtherFrame("etherFrame");

//    etherFrame->setTimestamp(msg->getTimestamp());

    // fill RAWEtherFrame message
    etherFrame->setByteLength(length);
    etherFrame->setRawBytesArraySize(length);
    char *bufferPtr = wlanPkt;
    for (int i = 0; i < length; i++)
    {
        etherFrame->setRawBytes(i, *bufferPtr);
        bufferPtr++;
    }
    etherFrame->setDest();

    // print some information
    MACAddress destAddr = (((Ieee802Ctrl *) etherFrame->getControlInfo())->getDest());
    ev << host->getName() << ":MsgHandler.sendToModel(): sending EtherFrame to " << destAddr.str() << "\n";

    // send via wireless stack (see VirtualHost.ned for details)
    send(etherFrame, "ifOut");
}

void MsgHandler::setRTS(int8_t index, int16_t rts)
{
	ev << host->getName() << ":MsgHandler.setRTS(" << rts << "B)\n";
	cModule *mac = check_and_cast<cModule *>(vif[index]->getSubmodule("mac"));
	mac->par("rtsThresholdBytes") = rts;
}

void MsgHandler::setRetry(int8_t index, int16_t retry)
{
	ev << host->getName() << ":MsgHandler.setRetry(" << retry << ")\n";
	cModule *mac = check_and_cast<cModule *>(vif[index]->getSubmodule("mac"));
	mac->par("retryLimit") = retry;
}

void MsgHandler::setTxPower(int8_t index, int16_t txpower)
{
	ev << host->getName() << ":MsgHandler.setTxPower(" << txpower << "mW)\n";
	cModule *radio = check_and_cast<cModule *>(vif[index]->getSubmodule("radio"));
	radio->par("transmitterPower") = txpower;
}

void MsgHandler::setSensitivity(int8_t index, int16_t sensitivity)
{
	ev << host->getName() << ":MsgHandler.setSensitivity(" << sensitivity << "dB)\n";
	cModule *radio = check_and_cast<cModule *>(vif[index]->getSubmodule("radio"));
	radio->par("sensitivity") = sensitivity;
}

void MsgHandler::updateVifParameter(configNotification *info)
{
	Enter_Method("updateVifParameter()");

	int vifIndex = 0; 	//TODO: set correct index
	if (vifIndex > numVif)
	{
		ev << "MsgHandler.updateVifParameter(): No interface with index " << vifIndex << "!\n";
		return;
	}
	switch(info->type)
	{
		case SIOCSIWFREQ:
			// TODO: set frequency
			break;
		case SIOCSIWRTS:
			setRTS(vifIndex, info->value.rts);
			break;
		case SIOCSIWRETRY:
			setRetry(vifIndex, info->value.retry);
			break;
		case SIOCSIWTXPOW:
			setTxPower(vifIndex, info->value.txpower);
			break;
		case SIOCSIWSENS:
			setSensitivity(vifIndex, info->value.sensitivity);
			break;
		default:
			ev << "MsgHandler.updateVifParameter(): Unknown ioctl() " << info->type << "\n";
			break;
	}
}

void MsgHandler::initializeVifParameters(vifRegistrationData vifReg[])
{
	Enter_Method("initializeVifParameters()");

	for (int i = 0; i < numVif; i++)
	{
		// TODO: set frequency
		setRTS(i, vifReg->rts);
		setRetry(i, vifReg->retry);
		setTxPower(i, vifReg->txpower);
		if (vifReg->sensitivity < 0)
			setSensitivity(i, vifReg->sensitivity);
	}
}
